package com.novaordis.thread194629;

import java.util.Collection;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.atomic.AtomicInteger;

import javax.transaction.TransactionManager;

import org.apache.log4j.Logger;
import org.infinispan.Cache;
import org.infinispan.configuration.cache.CacheMode;
import org.infinispan.configuration.cache.ConfigurationBuilder;
import org.infinispan.configuration.global.GlobalConfigurationBuilder;
import org.infinispan.manager.CacheManager;
import org.infinispan.manager.DefaultCacheManager;
import org.infinispan.transaction.LockingMode;
import org.infinispan.transaction.TransactionMode;
import org.infinispan.transaction.lookup.DummyTransactionManagerLookup;
import org.infinispan.tree.Node;
import org.infinispan.tree.TreeCache;
import org.infinispan.tree.TreeCacheFactory;


public class TestHelper {

   public static final int THREAD_COUNT = 15;
   public static final int KEY_COUNT = 1;

   static final TestHelper instance = new TestHelper();
   private ExecutorService executorService;
   private DefaultCacheManager infinispanCacheManager;
   private static final String DEFAULT_CLUSTER_NAME = "default-cluster";
   private final static int RETRIES_NUMBER = 5;
   private static final Logger logger = Logger.getLogger(TestHelper.class.getName());
   private ConcurrentHashMap<String, TestCacheProxy> caches = new ConcurrentHashMap<String, TestCacheProxy>();
   public static final String TREE_CACHE_SIGNATURE_KEY = "TreeCacheSignature";
   public static long LOCK_ACQUISITION_TIMEOUT = 300000L; // 5 min
   public static final int DEFAULT_CONCURRENCY_LEVEL = 2000;
   public static final long WAIT_FOR_CACHE_READY = 100;
   public static final String NO_MORE_CACHE_AVAILABLE_MSG = "No more caches available... Please release and retry";
   public static final String EXCEPTION_CREATE_SIGNED_CACHE = "Could not create or get a Signed Cache";
   public static long RETRY_CLUSTER_FORM_WAIT = 6000L;

   private static final int MAX_CACHES = 5;

   private TestHelper() {
      executorService = Executors.newFixedThreadPool(THREAD_COUNT);
     

      GlobalConfigurationBuilder gcb = GlobalConfigurationBuilder.defaultClusteredBuilder();
      org.infinispan.configuration.global.GlobalConfiguration gc = gcb.transport()
               .clusterName(DEFAULT_CLUSTER_NAME).build();

      ConfigurationBuilder cb = getDefaultClusteredCacheConfig(CacheMode.DIST_ASYNC, true);
      cb.transaction().lockingMode(LockingMode.PESSIMISTIC).locking().lockAcquisitionTimeout(100l)
               .invocationBatching().enable();

      infinispanCacheManager = new DefaultCacheManager(gc, cb.build());
      try {
         Thread.currentThread().sleep(5000);
      } catch (InterruptedException e) {
         // TODO Auto-generated catch block
         e.printStackTrace();
      }
      String[] cacheNames = getCacheNames(MAX_CACHES);
      startCaches(cb.build(), cacheNames);
   }

   public void startCaches(org.infinispan.configuration.cache.Configuration localConfiguration,
            String[] cacheNames) {
      // logger.info("startCaches start");

      for (String cacheName : cacheNames) {
         infinispanCacheManager.defineConfiguration(cacheName, localConfiguration);
         infinispanCacheManager.getCache(cacheName);

      }

      while (!infinispanCacheManager.getStatus().allowInvocations()) {
         try {
            Thread.sleep(WAIT_FOR_CACHE_READY);
         } catch (InterruptedException e) {

         }
      }

      if (infinispanCacheManager.getStatus().allowInvocations()) {
         for (String cacheName : cacheNames) {
            Cache<Object, Object> mapCache = infinispanCacheManager.getCache(cacheName);
            TreeCache<Object, Object> treeCache = getTreeCache(mapCache);
            TestCacheProxy tcp = new TestCacheProxy(mapCache, treeCache);
            caches.putIfAbsent(cacheName, tcp);
         }
      }
      if (null != cacheNames) {
         // logger.info("Created Caches " + cacheNames.length);
      }

   }

   public boolean isCacheSigned(final Cache<Object, Object> cache) {
      return getSignature(cache) != null;
   }

   public Cache<Object, Object> getAnyCache() throws Exception {
      Cache<Object, Object> availableCache = null;
      Collection<TestCacheProxy> valuesList = caches.values();

      for (TestCacheProxy cacheProxy : valuesList) {
         Cache<Object, Object> cache = (Cache<Object, Object>) cacheProxy.getMapCache();
         String cacheName = cache.getName();

         if (cacheName.equals("__defaultcache")) {
            continue;
         }
         if (!isCacheSigned(cache)) {
            availableCache = cache;
            break;
         }
      }
      if (availableCache == null) {
         throw new Exception(NO_MORE_CACHE_AVAILABLE_MSG);
      }

      return availableCache;
   }

   public Cache<Object, Object> createIfAbsent(ITreeCacheSignatureTest signature) throws Exception {

      Cache<Object, Object> dCache = getSignedCache(signature);
      if (dCache == null) {
         dCache = getAnyCache();
         // logger.info("createIfAbsent Got signed cache random " );
         putSignature(dCache, signature);
      }

      if (null == dCache) {
         throw new Exception(EXCEPTION_CREATE_SIGNED_CACHE);
      }
      return dCache;
   }

   public void putSignature(final Cache<Object, Object> cache,
            final ITreeCacheSignatureTest signature) throws Exception {
      if (signature == null) {
         throw new Exception("Cannot put a null signature");
      }

      putOnRoot(cache, TREE_CACHE_SIGNATURE_KEY, signature);
   }

   public static String[] getCacheNames(int maxNo) {
      AtomicInteger ai = new AtomicInteger();
      String[] cacheNames = new String[maxNo];
      for (int i = 0; i < maxNo; i++) {
         cacheNames[i] = "Cache-" + ai.getAndIncrement();
      }

      return cacheNames;
   }

   public static TestHelper getInstance() {
      return instance;
   }

   public ExecutorService getExecutorService() {
      return executorService;
   }

   public CacheManager getCacheManager() {
      return infinispanCacheManager;
   }

   public Cache<Object, Object> getSignedCache(final ITreeCacheSignatureTest signature)
            throws Exception {
      for (final TestCacheProxy cacheProxy : caches.values()) {
         Cache<Object, Object> cache = cacheProxy.getMapCache();
         if (cacheHasSignature(cache, signature)) {
            return cache;
         }
      }
      return null;

   }

   public boolean cacheHasSignature(final Cache<Object, Object> cache,
            final ITreeCacheSignatureTest signature) throws Exception {
      if (signature == null) {
         throw new Exception("Cannot evaluate a null signature");
      }

      final ITreeCacheSignatureTest availableSignature = getSignature(cache);
      return availableSignature == null ? false : availableSignature.equals(signature);
   }

   public TreeCache<Object, Object> getTreeCache(Cache<Object, Object> cache) {
      TreeCacheFactory treeFactory = new TreeCacheFactory();
      TreeCache<Object, Object> treeCache = treeFactory.createTreeCache(cache);
      return treeCache;
   }

   public ITreeCacheSignatureTest getSignature(final Cache<Object, Object> cache) {
      return getFromRoot(cache, TREE_CACHE_SIGNATURE_KEY);
   }

   public <T> T getFromRoot(final Cache<Object, Object> cache, final String key) {
      String cacheName = cache.getName();
      TestCacheProxy tcp = caches.get(cacheName);
      final TreeCache<Object, Object> treeCache = tcp.getTreeCache();
      if (treeCache != null) {
         final Node<Object, Object> rootNode = treeCache.getRoot();
         if (rootNode != null) {
            T signatureValue = null;
            int count = 0;
            while (count < RETRIES_NUMBER) {
               try {
                  signatureValue = (T) rootNode.get(key);
                  if (null != signatureValue) {
                     break;
                  } else {
                     // logger.info("Signature with Key: " + key
                     // + " could not be found for cache: " +
                     // getCacheName(cache));
                     break;
                  }
               } catch (Exception e) {
                  e.printStackTrace();
                  // logger.warn("Failed to get signature from root for cache"
                  // + cacheName + e);
               }
               // logger.warn("Retrying to get the signature from root - retrying "
               // + count);
               count++;
               try {
                  Thread.sleep(RETRY_CLUSTER_FORM_WAIT);
               } catch (InterruptedException ie) {
                  logger.warn("Waiting for another try on getting signature");

               }
            }
            return signatureValue;
         }

      }

      return null;
   }

   public <T> void putOnRoot(final Cache<Object, Object> cache, final String key, final T obj)
            throws Exception {
      String cacheName = cache.getName();
      TestCacheProxy tcp = caches.get(cacheName);
      final TreeCache<Object, Object> treeCache = tcp.getTreeCache();
      if (treeCache == null) {
         throw new Exception("TreeCache is null for cache: " + cacheName);
      } else {

         Node<Object, Object> rootNode = null;
         try {
            rootNode = treeCache.getRoot();
         } catch (Exception e) {
            e.printStackTrace();
            // logger.error(
            // "Failed getting root node on Cache " + cacheName, e);
         }

         if (rootNode != null) {
            TransactionManager tm = null;
            tm = cache.getAdvancedCache().getTransactionManager();
            try {
               logger.info("putOnRoot - about to begin transaction on Cache " + cacheName
                        + " for key: " + key);
               tm.begin();
               putOnClusterWithRetries(rootNode, key, obj);
               tm.commit();
               logger.info("putOnRoot - committed transaction on Cache " + cacheName + " for key: "
                        + key);
            } catch (Exception ex) {
               if (tm != null) {
                  try {
                     // logger.info("putOnRoot - rolling back transaction");
                     tm.rollback();
                  } catch (Exception e) {
                     logger.error("Unable to rollback", ex);
                  }
               }
               throw new Exception("Transaction failed putting on cluster", ex);
            }

         }

      }

   }

   private <T> void putOnClusterWithRetries(Node<Object, Object> rootNode, final String key,
            final T obj) {
      int count = 0;
      while (count <= RETRIES_NUMBER) {
         try {
            rootNode.put(key, obj);
            break;
         } catch (Exception e) {
            count++;
            // logger.warn("Failed to put Signature - retrying " + count);
            try {
               Thread.sleep(RETRY_CLUSTER_FORM_WAIT);
            } catch (InterruptedException ie) {
               // logger.warn("Waiting for another try on putting signature");

            }
         }
      }

   }
   private static void markAsTransactional(boolean transactional, ConfigurationBuilder builder) {
      builder.transaction().transactionMode(
               transactional ? TransactionMode.TRANSACTIONAL : TransactionMode.NON_TRANSACTIONAL);
   }

   private static void updateTransactionSupport(boolean transactional, ConfigurationBuilder builder) {
      if (transactional)
         amendJTA(builder);
   }

   private static void amendJTA(ConfigurationBuilder builder) {
      builder.transaction().transactionManagerLookup(new DummyTransactionManagerLookup());
   }

   public static ConfigurationBuilder getDefaultClusteredCacheConfig(CacheMode mode, boolean transactional) {
      ConfigurationBuilder builder = new ConfigurationBuilder();
      markAsTransactional(transactional, builder);
      updateTransactionSupport(transactional, builder);
      builder.clustering().cacheMode(mode).stateTransfer().fetchInMemoryState(false).transaction()
               .syncCommitPhase(true).syncRollbackPhase(true).cacheStopTimeout(0L);

      if (mode.isSynchronous())
         builder.clustering().sync();
      else
         builder.clustering().async();

      return builder;
   }
}
