package com.novaordis.thread194629;

import org.apache.log4j.Logger;
import org.infinispan.Cache;
import org.infinispan.config.Configuration;
import org.infinispan.config.GlobalConfiguration;
import org.infinispan.configuration.cache.LegacyConfigurationAdaptor;
import org.infinispan.configuration.global.LegacyGlobalConfigurationAdaptor;
import org.infinispan.manager.DefaultCacheManager;
import org.infinispan.manager.EmbeddedCacheManager;
import org.infinispan.util.concurrent.IsolationLevel;
import org.junit.Assert;
import org.junit.Test;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;

/**
 * @author <a href="mailto:ovidiu@novaordis.com">Ovidiu Feodorov</a>
 *
 * Copyright 2012 Nova Ordis LLC
 */
public class ConcurrentWriteReadTest extends Assert
{
    // Constants -----------------------------------------------------------------------------------

    private static final Logger log = Logger.getLogger(ConcurrentWriteReadTest.class);

    // Static --------------------------------------------------------------------------------------

    // Attributes ----------------------------------------------------------------------------------

    // Constructors --------------------------------------------------------------------------------

    // Public --------------------------------------------------------------------------------------

    @Test
    public void testConcurrentWriteRead() throws Exception
    {
        int threadCount = 15;
        int cacheEntryCount = 50;

        ExecutorService es = null;
        EmbeddedCacheManager cm = null;

        try
        {
            es = Executors.newFixedThreadPool(threadCount);

            cm = createAndConfigureCacheManager();

            Cache<Object, Object> c = cm.getCache("TEST-CACHE");

            // no need to start() the cache, creation implicitly starts

            while (!c.getCacheManager().getStatus().allowInvocations())
            {
                log.info("sleeping, waiting for the cache to start ...");
                Thread.sleep(500l);
            }

            List<Future> futures = new ArrayList<Future>();

            for(int i = 0; i < cacheEntryCount; i ++)
            {
                String s = "test" + (i < 10 ? "0" : "") + i;

                // submit asynchronous Runnables
                Future f = es.submit(new WriteAndRead(c, s));
                futures.add(f);
            }

            // wait for completion of all runnables - be aware they may also fail

            int completed = 0;
            int failed = 0;

            for(Future f: futures)
            {
                try
                {
                    f.get();
                    completed ++;
                }
                catch(Exception e)
                {
                    failed ++;
                }
            }

            log.info("completed: " + completed + ", failed: " + failed);

            // we expect zero failures
            assertEquals(0, failed);
        }
        finally
        {
            if (es != null)
            {
                es.shutdown();
            }

            if (cm != null)
            {
                cm.stop();
            }
        }
    }

    // Package protected ---------------------------------------------------------------------------

    // Protected -----------------------------------------------------------------------------------

    // Private -------------------------------------------------------------------------------------

    private EmbeddedCacheManager createAndConfigureCacheManager() throws Exception
    {
        GlobalConfiguration gc = GlobalConfiguration.getClusteredDefault();

        gc.setClusterName("TEST");

        Configuration lc = new Configuration();
        lc.setCacheMode(Configuration.CacheMode.DIST_ASYNC);
        lc.setRehashEnabled(true);
        lc.setSyncReplTimeout(300000L);
        lc.setLockAcquisitionTimeout(300000L);
        lc.setEnableDeadlockDetection(true);
        lc.setInvocationBatchingEnabled(true);
        lc.setEagerLockSingleNode(false);
        lc.setExpirationLifespan(-1);
        lc.setConcurrencyLevel(2000);
        lc.setIsolationLevel(IsolationLevel.READ_COMMITTED);
        lc.setSyncCommitPhase(true);
        lc.setSyncRollbackPhase(true);
        lc.setNumOwners(2);

        return new DefaultCacheManager(
            LegacyGlobalConfigurationAdaptor.adapt(gc),
            LegacyConfigurationAdaptor.adapt(lc), false);
    }

    // Inner classes -------------------------------------------------------------------------------
}
